@@ -1,26 +1,30 @@
 #include <stdio.h>
-#include "petsc.h"
-#include "mat.h"
-#include "pc.h"
+#include "./examples.h"
 
-int prec_setup(MPI_Comm comm,Mat A,PC *B)
+#undef __FUNC__
+#define __FUNC__ "prec_setup"
+int prec_setup(MPI_Comm comm,Mat A,PC the_pc)
 {
-  PC the_pc;
   PCType pc_name = PCDomainDecomp;
-  int mytid,ierr;
+  int ierr;
 
-/*  MPI_Comm_rank(comm,&mytid);*/
-  ierr = PCCreate(comm,&the_pc); CHKERRQ(ierr);
   /* ---------------------------------- *
    * ---- here is the crucial part ---- *
    * ---------------------------------- */
-/*  if (!mytid) printf("Setting up Domain Decomp pc %d\n",pc_name);*/
   ierr = PCSetType(the_pc,pc_name); CHKERRQ(ierr);
   /* ---------------------------------- */
   {
     PC intface_pc;
     ierr = PCDomainDecompGetInterfacePC(the_pc,&intface_pc); CHKERRQ(ierr);
-    ierr = PCSetType(intface_pc,PCJACOBI); CHKERRQ(ierr);
+    ierr = PCSetType(intface_pc,PCGenBlockSSOR); CHKERRQ(ierr);
+    {
+      PC very_local_pc;
+      ierr = PCParallelGetLocalPC(intface_pc,&very_local_pc); CHKERRQ(ierr);
+      ierr = PCSetType(very_local_pc,PCLU); CHKERRQ(ierr);
+    }
+    ierr = PCParallelSubdomainPipelineSetType
+              (intface_pc,PIPELINE_SEQUENTIAL);
+    CHKERRQ(ierr);
 /*
     ierr = PCSetType(intface_pc,PCAdditiveSchwarz); CHKERRQ(ierr);
     {
@@ -28,7 +32,8 @@ int prec_setup(MPI_Comm comm,Mat A,PC *B)
       ierr = PCParallelGetLocalPC(intface_pc,&very_local_pc); CHKERRQ(ierr);
       ierr = PCSetType(very_local_pc,PCLU); CHKERRQ(ierr);
     }
-    ierr = PCParallelSetCustomPipeline(intface_pc,PIPELINE_CUSTOM_SEQUENTIAL);
+    ierr = PCParallelSubdomainPipelineSetType
+              (intface_pc,PIPELINE_CUSTOM_SEQUENTIAL);
 */
   }
 
@@ -41,7 +46,5 @@ int prec_setup(MPI_Comm comm,Mat A,PC *B)
 
   ierr = ParPreSetup(comm,A,the_pc); CHKERRQ(ierr);
 
-  *B = the_pc;
-
   return 0;
 }
