@@ -1,16 +1,13 @@
 #include <stdio.h>
-#include "petsc.h"
-#include "mat.h"
-#include "pc.h"
-#include "parpre_pc.h"
+#include "./examples.h"
 
-int prec_setup(MPI_Comm comm,Mat A,PC *B)
+#undef __FUNC__
+#define __FUNC__ "prec_setup"
+int prec_setup(MPI_Comm comm,Mat A,PC the_pc)
 {
-  PC the_pc;
   int mytid,ierr;
 
   MPI_Comm_rank(comm,&mytid);
-  ierr = PCCreate(comm,&the_pc); CHKERRQ(ierr);
 
   if (!mytid) printf("Setting up Generalised Block SSOR pc\n");
   /* ---------------------------------- *
@@ -22,19 +19,17 @@ int prec_setup(MPI_Comm comm,Mat A,PC *B)
   {
     PC local_pc;
     ierr = PCParallelGetLocalPC(the_pc,&local_pc); CHKERRQ(ierr);
-    ierr = PCSetType(local_pc,PCILU); CHKERRQ(ierr);
+    ierr = PCSetType(local_pc,PCLU); CHKERRQ(ierr);
   }
 
-  ierr = PCGenBlockSSORSetNoGlobalFactorisation(the_pc); CHKERRQ(ierr);
-  ierr = PCParallelSetCustomPipeline(the_pc, PIPELINE_CUSTOM_SEQUENTIAL );
+  ierr = PCGenBlockSSORSetGlobalFactorisation(the_pc); CHKERRQ(ierr);
+  ierr = PCParallelSubdomainPipelineSetType(the_pc,PIPELINE_SEQUENTIAL);
   CHKERRQ(ierr);
 
   /* ---------------------------------- */
 
   ierr = ParPreSetup(comm,A,the_pc); CHKERRQ(ierr);
 
-  *B = the_pc;
-
   return 0;
 }
 
