@@ -1,16 +1,15 @@
 #include <stdio.h>
-#include "petsc.h"
-#include "mat.h"
-#include "pc.h"
+#include "./examples.h"
 
-int prec_setup(MPI_Comm comm,Mat A,PC *B)
+#undef __FUNC__
+#define __FUNC__ "prec_setup"
+int prec_setup(MPI_Comm comm,Mat A,PC the_pc)
 {
-  PC the_pc;
   PCType pc_name = PCJACOBI;
   int mytid,ierr;
 
   MPI_Comm_rank(comm,&mytid);
-  ierr = PCCreate(comm,&the_pc); CHKERRQ(ierr);
+
   /* ---------------------------------- *
    * ---- here is the crucial part ---- *
    * ---------------------------------- */
@@ -20,8 +19,6 @@ int prec_setup(MPI_Comm comm,Mat A,PC *B)
 
   ierr = ParPreSetup(comm,A,the_pc); CHKERRQ(ierr);
 
-  *B = the_pc;
-
   return 0;
 }
 
