@@ -1,19 +1,17 @@
 #include <stdio.h>
-#include "petsc.h"
-#include "mat.h"
-#include "pc.h"
-#include "parpre_pc.h"
+#include "./examples.h"
 
-int prec_setup(MPI_Comm comm,Mat A,PC *B)
+#undef __FUNC__
+#define __FUNC__ "prec_setup"
+int prec_setup(MPI_Comm comm,Mat A,PC the_pc)
 {
-  PC the_pc;
   PCType pc_name = PCMultiplicativeSchwarz;
   int mytid,ierr;
 
   MPI_Comm_rank(comm,&mytid);
-  ierr = PCCreate(comm,&the_pc); CHKERRQ(ierr);
 
-  if (!mytid) printf("Setting up Multiplicative Schwarz pc %d\n",pc_name);
+  if (!mytid) printf("Setting up Multiplicative Schwarz pc %d\n",
+		     (int)pc_name);
   /* ---------------------------------- *
    * ---- here is the crucial part ---- *
    * ---------------------------------- */
@@ -23,7 +21,7 @@ int prec_setup(MPI_Comm comm,Mat A,PC *B)
   ierr = PCSchwarzSetHaloSize(the_pc,1); CHKERRQ(ierr);
 
   /* Declare global communication scheme */
-  ierr = PCParallelSetCustomPipeline(the_pc,PIPELINE_CUSTOM_REDBLACK);
+  ierr = PCParallelSubdomainPipelineSetType(the_pc,PIPELINE_SEQUENTIAL);
 
   /* Declare local solve */
   {
@@ -39,8 +37,6 @@ int prec_setup(MPI_Comm comm,Mat A,PC *B)
 
   ierr = ParPreSetup(comm,A,the_pc); CHKERRQ(ierr);
 
-  *B = the_pc;
-
   return 0;
 }
 
