@@ -1,11 +1,11 @@
 #include "src/mat/impls/aij/mpi/mpiaij.h"
 #include "src/vec/vecimpl.h"
-#include "src/pc/pcextra.h"
-#include "src/vec/utils/vpipe.h"
+#include "parpre_pipeline.h"
+#include "src/vec/utils/pipeline_impl.h"
+#include "parpre_mat.h"
 #include "parpre_pc.h"
-#include "./mpixtra.h"
-
-extern int MatMatMult_AIJ(Mat a, Mat b, Mat *c);
+#include "src/mat/parpre_matimpl.h"
+#include "src/pc/pcextra.h"
 
 /* VE functions added from here on down */
 int MatGetRowLen_MPIAIJ(Mat mat,int row,int *len)
@@ -14,12 +14,13 @@ int MatGetRowLen_MPIAIJ(Mat mat,int row,int *len)
   Mat dia = Aij->A, off = Aij->B;
   Mat_SeqAIJ *aij = (Mat_SeqAIJ *) dia->data, *bij = (Mat_SeqAIJ *) off->data;
 
-  if (row<Aij->rstart | row>=Aij->rend)
+  if ( (row<Aij->rstart) | (row>=Aij->rend) )
     SETERRQ(1,0,"GetRowLen only for local rows");
   row -= Aij->rstart;
   *len = aij->i[row+1] - aij->i[row] + bij->i[row+1] - bij->i[row];
   return 0;
 }
+
 int MatMaxRowLen_MPIAIJ(Mat A, int *rowlen)
 {
   Mat_MPIAIJ *Aij = (Mat_MPIAIJ *) A->data;
@@ -82,8 +83,8 @@ int MatMatMult_MPIAIJ(Mat A,Mat B,Mat *C)
 {
   Mat_MPIAIJ  *Aij = (Mat_MPIAIJ *) A->data;
   MPI_Comm    comm = A->comm;
-  Mat *sub,tmp1,tmp2,res;
-  int iRow,ierr;
+  Mat res;
+  int ierr;
   int A_start,A_local,A_global,B_global;
   int A_off = ((Mat_SeqAIJ*) Aij->B->data)->n;
 
@@ -100,22 +101,20 @@ int MatMatMult_MPIAIJ(Mat A,Mat B,Mat *C)
     ierr = MatGetOwnershipRange(A,&A_start,&idum); CHKERRQ(ierr);
     ierr = MatGetLocalSize(A,&A_local,&A_jlocal); CHKERRQ(ierr);
     ierr = MatGetLocalSize(B,&B_local,&B_jlocal); CHKERRQ(ierr);
-    if (A_global==B_global) {
-      ierr = MatCreateMPIAIJ(comm,A_local,A_local,PETSC_DECIDE,PETSC_DECIDE,
-			     0,0,0,0,&res); CHKERRQ(ierr);
-    } else {/* is this case ever appropriate? */
+    if (A_global!=B_global) {
+      /* is this case ever appropriate? */
+printf(">>>> Strange case: MatMatMult_MPIAIJ <<<<\n");
       ierr = MatCreateMPIAIJ(comm,PETSC_DECIDE,PETSC_DECIDE,A_global,B_global,
 			     0,0,0,0,&res); CHKERRQ(ierr);
     }
   }
 
-  /* Get the two subblocks from B */
-  sub = (Mat *) PetscMalloc(2*sizeof(Mat)); CHKPTRQ(sub);
   {
-    IS A_home,A_away,B_total,*IS_A,*IS_B;
+    Mat *sub,tmp1,tmp2;
+    IS A_home,A_away,B_total,IS_A[2],IS_B[2];
+    int *band, iRow;
+
     /* index sets in A and B */
-    IS_A = (IS *) PetscMalloc(2*sizeof(IS)); CHKPTRQ(IS_A);
-    IS_B = (IS *) PetscMalloc(2*sizeof(IS)); CHKPTRQ(IS_B);
     ierr = ISCreateStride(comm,B_global,0,1,&B_total); CHKERRQ(ierr);
     ierr = ISCreateGeneral(comm,A_off,Aij->garray,&A_away); CHKERRQ(ierr);
     ierr = ISCreateStride(comm,Aij->cend-Aij->cstart,Aij->cstart,1,&A_home);
@@ -126,15 +125,38 @@ int MatMatMult_MPIAIJ(Mat A,Mat B,Mat *C)
     ierr = MatGetSubMatrices(B,2,IS_A,IS_B,MAT_INITIAL_MATRIX,&sub);
     CHKERRQ(ierr);
     ISDestroy(A_home); ISDestroy(A_away); ISDestroy(B_total);
-    PetscFree(IS_A); PetscFree(IS_B);
+
+    /* multiply diag and off-diag block of A with relevant parts of B */
     {
       Mat diag_factor = sub[0], off_factor = sub[1];
       ierr = MatMatMult_AIJ(Aij->A,diag_factor,&tmp1); CHKERRQ(ierr);
       ierr = MatDestroy(diag_factor); CHKERRQ(ierr);
       ierr = MatMatMult_AIJ(Aij->B,off_factor,&tmp2); CHKERRQ(ierr);
       ierr = MatDestroy(off_factor); CHKERRQ(ierr);
-      PetscFree(sub);
     }
+    PetscFree(sub);
+    /* (over)estimate resulting bandwidth */
+    band = (int *) PetscMalloc((A_local+1)*sizeof(int)); CHKPTRQ(band);
+    for (iRow=0; iRow<A_local; iRow++) {
+      int ncols;
+      ierr = MatGetRow(tmp1,iRow,&ncols,PETSC_NULL,PETSC_NULL);
+      CHKERRQ(ierr);
+      band[iRow] = ncols;
+      ierr = MatRestoreRow(tmp1,iRow,&ncols,PETSC_NULL,PETSC_NULL);
+      CHKERRQ(ierr);
+      
+      ierr = MatGetRow(tmp2,iRow,&ncols,PETSC_NULL,PETSC_NULL);
+      CHKERRQ(ierr);
+      band[iRow] += ncols;
+      ierr = MatRestoreRow(tmp2,iRow,&ncols,PETSC_NULL,PETSC_NULL);
+      CHKERRQ(ierr);
+    }
+    /* create matrix and fill in */
+    ierr = MatCreateMPIAIJ
+      (comm,A_local,A_local,PETSC_DECIDE,PETSC_DECIDE,
+       PETSC_NULL,band,PETSC_NULL,band,&res);
+    CHKERRQ(ierr);
+    PetscFree(band);
     for (iRow=0; iRow<A_local; iRow++) {
       int Row=A_start+iRow,ncols,*cols; Scalar *vals;
 
@@ -148,17 +170,136 @@ int MatMatMult_MPIAIJ(Mat A,Mat B,Mat *C)
       CHKERRQ(ierr);
       ierr = MatRestoreRow(tmp2,iRow,&ncols,&cols,&vals); CHKERRQ(ierr);
     }
+    ierr = MatDestroy(tmp1); CHKERRQ(ierr);
+    ierr = MatDestroy(tmp2); CHKERRQ(ierr);
   }
   ierr = MatAssemblyBegin(res,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
   ierr = MatAssemblyEnd(res,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
-  ierr = MatDestroy(tmp1); CHKERRQ(ierr);
-  ierr = MatDestroy(tmp2); CHKERRQ(ierr);
 
   *C = res;
 
   return 0;
 }
 
+#undef __FUNC__
+#define __FUNC__ "MatrixAij2MpiAbut"
+int MatrixAij2MpiAbut(int part,Mat in,MPI_Comm comm,Mat *out)
+{
+  Mat res;
+  int *band,m,n,irow,rstart,rend,ierr;
+
+  /* get the bandwidth of the matrix to be embedded,
+     specifying this as d_nnz/o_nnz is an overestimate */
+  ierr = MatGetLocalSize(in,&m,&n); CHKERRQ(ierr);
+  band = (int *) PetscMalloc((m+1)*sizeof(int)); CHKPTRQ(band);
+  for (irow=0; irow<m; irow++) {
+    ierr = MatGetRow(in,irow,band+irow,PETSC_NULL,PETSC_NULL);
+    CHKERRQ(ierr);
+    ierr = MatRestoreRow(in,irow,band+irow,PETSC_NULL,PETSC_NULL);
+    CHKERRQ(ierr);
+  }
+
+  ierr = MatCreateMPIAIJ(comm,m,part,PETSC_DECIDE,PETSC_DECIDE,
+			 PETSC_NULL,band,PETSC_NULL,band,&res);
+  CHKERRQ(ierr);
+  PetscFree(band);
+
+  ierr = MatGetOwnershipRange(res,&rstart,&rend); CHKERRQ(ierr);
+  for (irow=0; irow<m; irow++) {
+    int row=rstart+irow,ncols,*cols; Scalar *vals;
+    ierr = MatGetRow(in,irow,&ncols,&cols,&vals); CHKERRQ(ierr);
+    ierr = MatSetValues(res,1,&row,ncols,cols,vals,INSERT_VALUES); CHKERRQ(ierr);
+    ierr = MatRestoreRow(in,irow,&ncols,&cols,&vals); CHKERRQ(ierr);
+  }
+  ierr = MatAssemblyBegin(res,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
+  ierr = MatAssemblyEnd(res,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
+
+  *out = res;
+  return 0;
+}
+
+#undef __FUNC__
+#define __FUNC__ "MatrixAij2MpiIndex"
+int MatrixAij2MpiIndex(IS ptr,int m,int n,int M,int N,
+		       Mat in,MPI_Comm comm,Mat *out)
+{
+  Mat res;
+  int irow,ierr, nidx,*idx;
+
+printf(">> Using unoptimised routine Aij2MpiIndex <<\n");
+  ierr = ISGetSize(ptr,&nidx); CHKERRQ(ierr);
+  {
+    int ml,nl;
+    ierr = MatGetLocalSize(in,&ml,&nl); CHKERRQ(ierr);
+    if (m!=nidx) SETERRQ(1,0,"MatrixAij2MpiIndex: size mismatch");
+  }
+
+  ierr = MatCreateMPIAIJ(comm,m,n,M,N,5,PETSC_NULL,3,PETSC_NULL,&res);
+  CHKERRQ(ierr);
+
+  ierr = ISGetIndices(ptr,&idx); CHKERRQ(ierr);
+  for (irow=0; irow<m; irow++) {
+    int row=idx[irow],ncols,*cols; Scalar *vals;
+    ierr = MatGetRow(in,irow,&ncols,&cols,&vals); CHKERRQ(ierr);
+    ierr = MatSetValues(res,1,&row,ncols,cols,vals,INSERT_VALUES); CHKERRQ(ierr);
+    ierr = MatRestoreRow(in,irow,&ncols,&cols,&vals); CHKERRQ(ierr);
+  }
+  ierr = MatAssemblyBegin(res,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
+  ierr = MatAssemblyEnd(res,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
+
+  *out = res;
+  return 0;
+}
+
+/****************************************************************
+ * Matrix Scatter Pipelines                                     *
+ ****************************************************************/
+#undef __FUNC__
+#define __FUNC__ "MatCreateScatterPipeline"
+int MatCreateScatterPipeline
+(Mat base_mat,PipelineType pipe_type,
+ Vec *border_vec,VecPipeline *main_pipe)
+{
+  Mat_MPIAIJ *Aij = (Mat_MPIAIJ *) base_mat->data;
+  MPI_Comm comm = base_mat->comm;
+  VecPipeline pipe;
+  Vec tmp_g,bvec; IS is_g,is_l;
+  int local_size,low_border,local_border,local_domain;
+  int dum,ierr;
+
+  ierr = MatGetOwnershipRange(base_mat,&local_domain,&dum); CHKERRQ(ierr);
+  local_size = dum-local_domain;
+  ierr = VecCreateMPI(comm,local_size,PETSC_DECIDE,&tmp_g); CHKERRQ(ierr);
+  ierr = VecAssemblyBegin(tmp_g); CHKERRA(ierr);
+  ierr = VecAssemblyEnd(tmp_g); CHKERRA(ierr);
+  local_border= ((Mat_SeqAIJ *)Aij->B->data)->n;
+  ierr = ISCreateGeneral(MPI_COMM_SELF,local_border,Aij->garray,&is_g);
+  CHKERRQ(ierr);
+
+  ierr = VecCreateMPI
+    (comm,local_border,PETSC_DECIDE,&bvec); CHKERRQ(ierr);
+  ierr = VecAssemblyBegin(bvec); CHKERRA(ierr);
+  ierr = VecAssemblyEnd(bvec); CHKERRA(ierr);
+  ierr = VecGetOwnershipRange(bvec,&low_border,&dum); CHKERRQ(ierr);
+  ierr = ISCreateStride
+    (MPI_COMM_SELF,local_border,low_border,1,&is_l);
+  CHKERRQ(ierr);
+  
+  ierr = VecPipelineCreate(comm,tmp_g,is_g,bvec,is_l,&pipe);
+  CHKERRQ(ierr);
+  ierr = VecPipelineSetType(pipe,pipe_type,(PetscObject)base_mat);
+  CHKERRQ(ierr);
+  ierr = VecPipelineSetup(pipe); CHKERRQ(ierr);
+
+  ierr = VecDestroy(tmp_g); CHKERRQ(ierr);
+  ierr = ISDestroy(is_g); CHKERRQ(ierr);
+  ierr = ISDestroy(is_l); CHKERRQ(ierr);
+
+  *border_vec = bvec; *main_pipe = pipe;
+
+  return 0;
+}
+  
 /* VE from here on down probably superfluous */
 #define IBLOCK_TAG 11
 #define RBLOCK_TAG 12
@@ -219,13 +360,15 @@ int MatGatherCtxCreate(Mat mat,IS wanted,MatGatherCtx *rgs)
   for (i=p; i<=numtids; i++) request_row_ptrs[i] = request_nrows_tot;
   for (p=0; p<numtids; p++)
     request_nrows[p] = request_row_ptrs[p+1]-request_row_ptrs[p];
-/*
+
+  /*
 printf("I am requesting: ");
 for (i=0; i<numtids; i++) {int j;
 printf("(%d:) ",i); for (j=request_row_ptrs[i];j<request_row_ptrs[i+1];j++)
 printf("%d ",request_rows[j]);}
 printf("\n");
-*/
+  */
+
   /* every processor gathers how much is wanted from it from each proc */
   send_nrows = (int *) PetscMalloc( (numtids+1)*sizeof(int) );
   CHKPTRQ(send_nrows); PetscMemzero(send_nrows,(numtids+1)*sizeof(int));
@@ -276,15 +419,17 @@ printf("\n");
        (void *)send_rows,send_nrows,send_rows_ptrs,MPI_INT,p,comm);
     CHKERRQ(ierr);
   }
-/*
+
+  /*
 printf("I am sending: ");
 for (i=0; i<numtids; i++) {int j;
 printf("(%d:) ",i); for (j=send_rows_ptrs[i];j<send_rows_ptrs[i+1];j++)
 printf("%d ",send_rows[j]);}
 printf("\n");
-*/
+  */
+
   /* save reusable information */
-  gs = PetscNew(_MatGatherCtx);
+  gs = (MatGatherCtx) PetscMalloc(sizeof(struct _MatGatherCtx)); CHKPTRQ(gs);
   ierr = VecScatterCopy(Aij->Mvctx,(VecScatter*)&(gs->vs)); CHKERRQ(ierr);
   gs->comm = comm; gs->N = Aij->N;
   gs->mytid = mytid; gs->numtids = numtids;
@@ -327,25 +472,38 @@ static int MatGatherRowsReceive
 
   if (pf) gs->custom_recvs = 1;
   else gs->custom_recvs = 0;
-  
+
+  /*  
+printf("Selecting sender procs out of: ");
+for(ip=0;ip<gs->nrecvs_all;ip++)printf("%d, ",gs->recv_procs_all[ip]);printf("\n");
+printf("With");if((int)pf);else printf("out");printf(" selection function\n");
+  */
+
   recv_procs = (int *) PetscMalloc( (gs->nrecvs_all+1)*sizeof(int) );
   CHKPTRQ(recv_procs);
   nrecvs = 0;
   for (ip=0; ip<gs->nrecvs_all; ip++) {
     int p = gs->recv_procs_all[ip];
-/*    if (p==gs->mytid) continue;*/
-    if (pf && !(*pf)(p,pl)) ;
-    else
+    /*    if (p==gs->mytid) continue;*/
+    if (pf) 
+      if ((*pf)(p,pl)) {
+	recv_procs[nrecvs++] = p;
+      } else {
+      }
+    else {
       recv_procs[nrecvs++] = p;
+    }
   }
-
+  /*
+printf("Setting up to receive from: ");
+for(ip=0;ip<nrecvs;ip++)printf("%d, ",recv_procs[ip]);printf("\n");
+  */
   ierr = MatCreateSeqAIJ
     (MPI_COMM_SELF,gs->request_nrows_tot,gs->N,0,0,&catch_mat);
   CHKERRQ(ierr);
 
   count = nrecvs;
   while (count) {
-    int p,buflen,nrows,irow,unpacloc=0; void *buf;
     MPI_Status recv_status;
     while (1)
       for (ip=0; ip<nrecvs; ip++) {
@@ -356,36 +514,39 @@ static int MatGatherRowsReceive
 	if (flag) goto found;
       }
   found:
-    p = recv_status.MPI_SOURCE;
-    MPI_Get_count(&recv_status,MPI_BYTE,&buflen);
-    buf = (void *) PetscMalloc( buflen*sizeof(MPI_BYTE) );
-    CHKPTRQ(buf);
-    ierr = MPI_Recv(buf,buflen,MPI_PACKED,p,MBLOCK_TAG,comm,&recv_status);
-    CHKERRQ(ierr);
-    nrows = gs->request_row_ptrs[p+1]-gs->request_row_ptrs[p];
     {
-      int nrows_test;
-      MPI_Unpack(buf,buflen,&unpacloc,&nrows_test,1,MPI_INT,comm);
-      if (!(nrows==nrows_test)) SETERRQ(p+1,0,"Nrows mismatch");
+      int p,buflen,nrows,irow,unpacloc=0; void *buf;
+      
+      p = recv_status.MPI_SOURCE;
+      MPI_Get_count(&recv_status,MPI_BYTE,&buflen);
+      buf = (void *) PetscMalloc( buflen*sizeof(MPI_BYTE) );
+      CHKPTRQ(buf);
+      MPI_Recv
+	(buf,buflen,MPI_PACKED,p,MBLOCK_TAG,comm,&recv_status);
+      nrows = gs->request_row_ptrs[p+1]-gs->request_row_ptrs[p];
+      {
+	int nrows_test;
+	MPI_Unpack(buf,buflen,&unpacloc,&nrows_test,1,MPI_INT,comm);
+	if (!(nrows==nrows_test)) SETERRQ(p+1,0,"Nrows mismatch");
+      }
+      for (irow=0; irow<nrows; irow++) {
+	int rlen,*idx, row=gs->request_row_ptrs[p]+irow;
+	Scalar *val;
+	MPI_Unpack(buf,buflen,&unpacloc,&rlen,1,MPI_INT,comm);
+	idx = (int *) PetscMalloc( (1+rlen)*sizeof(int) );
+	CHKPTRQ(idx);
+	MPI_Unpack(buf,buflen,&unpacloc,idx,rlen,MPI_INT,comm);
+	val = (Scalar *) PetscMalloc( (1+rlen)*sizeof(Scalar) );
+	CHKPTRQ(val);
+	MPI_Unpack(buf,buflen,&unpacloc,val,rlen,MPI_DOUBLE,comm);
+	ierr = MatSetValues(catch_mat,1,&row,rlen,idx,val,INSERT_VALUES);
+	CHKERRQ(ierr);
+	PetscFree(idx); PetscFree(val);
+      }
+      PetscFree(buf);
     }
-    for (irow=0; irow<nrows; irow++) {
-      int rlen,*idx, row=gs->request_row_ptrs[p]+irow;
-      Scalar *val;
-      MPI_Unpack(buf,buflen,&unpacloc,&rlen,1,MPI_INT,comm);
-      idx = (int *) PetscMalloc( (1+rlen)*sizeof(int) );
-      CHKPTRQ(idx);
-      MPI_Unpack(buf,buflen,&unpacloc,idx,rlen,MPI_INT,comm);
-      val = (Scalar *) PetscMalloc( (1+rlen)*sizeof(Scalar) );
-      CHKPTRQ(val);
-      MPI_Unpack(buf,buflen,&unpacloc,val,rlen,MPI_DOUBLE,comm);
-      ierr = MatSetValues(catch_mat,1,&row,rlen,idx,val,INSERT_VALUES);
-      CHKERRQ(ierr);
-      PetscFree(idx); PetscFree(val);
-    }
-    PetscFree(buf);
     count--;
   }
-
   ierr = MatAssemblyBegin(catch_mat,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
   ierr = MatAssemblyEnd(catch_mat,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
 
@@ -400,7 +561,6 @@ static int MatGatherRowsReceive
     PetscFree(gs->send_buffers);
     gs->nsends_outstanding = 0;
   }
-
   
   *return_mat = catch_mat;
 
@@ -419,6 +579,12 @@ static int MatGatherRowsSend
   if (!pf) gs->custom_sends = 0;
   else gs->custom_sends = 1;
   
+  /*
+printf("Selecting receiver procs out of: ");
+for(ip=0;ip<gs->nsends_all;ip++)printf("%d, ",gs->send_procs_all[ip]);printf("\n");
+printf("With");if((int)pf);else printf("out");printf(" selection function\n");
+  */
+
   /* Determine what processors we are going to send to */
   send_procs = (int *) PetscMalloc( (gs->nsends_all+1)*sizeof(int) );
   CHKPTRQ(send_procs);
@@ -426,11 +592,20 @@ static int MatGatherRowsSend
   for (ip=0; ip<gs->nsends_all; ip++) {
     int p = gs->send_procs_all[ip];
 /*    if (p==gs->mytid) continue;*/
-    if (pf && !(*pf)(p,pl)) ;
-    else 
+    if (pf)
+      if ((*pf)(p,pl)) {
+	send_procs[nsends++] = p;
+      } else {
+      }
+    else {
       send_procs[nsends++] = p;
+    }
   }
 
+  /*
+printf("Setting up to send to: ");
+for(ip=0;ip<nsends;ip++)printf("%d, ",send_procs[ip]);printf("\n");
+  */
   if ((!gs->custom_sends) && nsends) {
     gs->nsends_outstanding = nsends;
     if (nsends) {
@@ -486,14 +661,13 @@ static int MatGatherRowsSend
     if (pacloc>s) SETERRQ(ip+1,0,"Pack overflow");
 
     if (gs->custom_sends) {
-      ierr = MPI_Send
+      MPI_Send
 	(packed_send_buffer,pacloc,MPI_PACKED,p,MBLOCK_TAG,comm);
-      CHKERRQ(ierr);
       PetscFree(packed_send_buffer);
     } else {
-      ierr = MPI_Isend
+      MPI_Isend
 	(packed_send_buffer,pacloc,MPI_PACKED,p,MBLOCK_TAG,comm,
-	 gs->send_requests+ip); CHKERRQ(ierr);
+	 gs->send_requests+ip);
       gs->send_buffers[ip] = packed_send_buffer;
     }
   }
@@ -514,43 +688,48 @@ int MatGatherRows(Mat mat,MatGatherCtx gs, Mat *res_mat)
 }
 
 int MatGatherRowsPipelineBegin
-(Mat mat,PipelineMode ptype, MatGatherCtx gs, Mat *ret_mat)
+(Mat mat,PipelineDirection ptype, MatGatherCtx gs, Mat *ret_mat)
 {
   VecPipeline pipe = gs->vs; PipelineFunction pf;
-  PetscObject pl = pipe->custom_pipe_data;
   int ierr;
 
-  if (ptype == PIPELINE_CUSTOM_UP)
-    pf = pipe->dn_fun;
+  if (ptype == PIPELINE_UP)
+    pf = pipe->dnfn;
   else
-    pf = pipe->up_fun;
-  if (!pf) {
+    pf = pipe->upfn;
+
+  if (!(int)pf || IsProcYes((int)pf) ) {
     ierr = MatGatherRowsSend(gs,mat,0,0); CHKERRQ(ierr);
     ierr = MatGatherRowsReceive(gs,0,0,ret_mat);
   } else {
+    PetscObject pl = pipe->custom_pipe_data;
     ierr = MatGatherRowsReceive(gs,pf,pl,ret_mat);
   }
   return 0;
 }
 
 int MatGatherRowsPipelineEnd
-(Mat mat,PipelineMode ptype, MatGatherCtx gs, Mat *ret_mat)
+(Mat mat,PipelineDirection ptype, MatGatherCtx gs, Mat *ret_mat)
 {
   VecPipeline pipe = gs->vs; PipelineFunction pf;
   PetscObject pl = pipe->custom_pipe_data;
-  int ierr;
+  int ierr,flg;
 
-  if (ptype == PIPELINE_CUSTOM_UP)
-    pf = pipe->up_fun;
+  if (ptype == PIPELINE_UP)
+    pf = pipe->upfn;
   else
-    pf = pipe->dn_fun;
+    pf = pipe->dnfn;
 
-  if (pf) {ierr = MatGatherRowsSend(gs,mat,pf,pl); CHKERRQ(ierr);}
+  flg = IsProcYes((int)pf);
+
+  if (((int)pf) && (!flg))
+    {ierr = MatGatherRowsSend(gs,mat,pf,pl); CHKERRQ(ierr);}
 
   return 0;
 }
 
 /* VE !!! not removed */
+/*
 int MatGetSubMatrix_MPIAIJ(Mat A,IS isrow,IS iscol,MatGetSubMatrixCall scall,Mat *B)
 {
   MatGatherCtx get_strip;
@@ -579,3 +758,4 @@ int MatGetSubMatrix_MPIAIJ(Mat A,IS isrow,IS iscol,MatGetSubMatrixCall scall,Mat
 
   return 0;
 }
+*/
