@@ -1,18 +1,14 @@
-#ifndef lint
-static char vcid[] = "$Id: precon.c,v 1.83 1996/04/09 20:24:08 curfman Exp bsmith $";
-#endif
 /*
     The PC (preconditioner) interface routines, callable by users.
 */
 #include "pc.h" /* VE this shoudln't be, but it gets taken from petsc, not parpre */
 #include "src/pc/pcimpl.h"            /*I "pc.h" I*/
-#include "pinclude/pviewer.h"
 
 #undef __FUNC__
 #define __FUNC__ "PCIsParallel"
 int PCIsParallel(PC pc) /* VE */
 {
-  if (pc->type>=PCAdditiveSchwarz) return 1;
+  if (pc->type >= PCAdditiveSchwarz) return 1;
   else return 0;
 }
 
@@ -20,7 +16,7 @@ int PCIsParallel(PC pc) /* VE */
 #define __FUNC__ "PCIsSchwarzMethod"
 int PCIsSchwarzMethod(PC pc) /* VE */
 {
-  if (pc->type==PCAdditiveSchwarz | pc->type==PCMultiplicativeSchwarz) 
+  if ((pc->type==PCAdditiveSchwarz) | (pc->type==PCMultiplicativeSchwarz))
     return 1;
   else return 0;
 }
